% This is a function to return a cell of all the trees that can be formed 
% from a base tree by pruning a single non-leaf node. 
%
% Usage: dTrees = pruneAllNodes(dT)
%
% Inputs: 
%
%   dT - the base decision tree to start with
%
% Outputs: 
%
%   dTrees - a cell of length T, where T is the number of possible trees
%   that can be fromed by pruning a single node from dT.  Each entry in
%   dTrees is a new decision tree. 
%
% NOTE: The trees returned by this function will be of the exact same
% structure as defined for the ouput of the function trainDT().  However, 
% following paths to leaf nodes may no longer result in a "pure"
% classificaiton of the training data. 
%
function dTrees = pruneAllNodes(dT)

if dT.isLeaf
    % If the node is already a leaf, simply return it. 
    dTrees = {dT};
else
    
    %** Construct new leaf node 
    [blah, newInd] = max(dT.trainData(2,:));
    newLabel = dT.trainData(1, newInd);
    newTree.isLeaf = true;
    newTree.label = newLabel;
    newTree.trainData = dT.trainData;
    
    % Get all pruned trees from child1
    if ~dT.child1.isLeaf
        child1Trees = pruneAllNodes(dT.child1);     
        nChild1Trees = length(child1Trees); 
    else
        nChild1Trees = 0; 
    end
    
    % Get all pruned trees from child2
    if ~dT.child2.isLeaf
        child2Trees = pruneAllNodes(dT.child2);
        nChild2Trees = length(child2Trees); 
    else
        nChild2Trees = 0; 
    end
    
    % Put everything together
    nNewTrees = 1 + nChild1Trees + nChild2Trees;
    dTrees = cell(1, nNewTrees); 
   
    dTrees{1} = newTree; 
    curPos = 1; 
    
    % Add all possible trees formed by pruning a single node from child 1
    baseTree = dT;
    for t = 1:nChild1Trees
        curPos = curPos + 1;
        curChild1Tree = child1Trees{t};
        
        if dT.child2.isLeaf && curChild1Tree.isLeaf
            % If child2 is a leaf, we need to check for the special case
            % that the reduced child1 is also a leaf with the same label as
            % the leaf on child2.  In this case, we don't want to add this
            % tree
            if curChild1Tree.label ~= dT.child2.label
                baseTree.child1 = curChild1Tree;
                dTrees{curPos} = baseTree;
            end
        else
            baseTree.child1 = curChild1Tree;
            dTrees{curPos} = baseTree;
        end
        
        
    end

    % Add all possible trees formed by pruning a single node from child 2
    baseTree = dT; 
    for t = 1:nChild2Trees
        curPos = curPos + 1;
        curChild2Tree = child2Trees{t};
        
        if dT.child1.isLeaf && curChild2Tree.isLeaf
            % Check for special case of matching leaf trees - see note in
            % code for child1Trees above
            if curChild2Tree.label ~= dT.child1.label
                baseTree.child2 = curChild2Tree;
                dTrees{curPos} = baseTree;
            end
        else
            baseTree.child2 = curChild2Tree;
            dTrees{curPos} = baseTree;
        end
    end
    
    % Clean up dTrees by removing empty entries we did not use
    emptyInds = cellfun('isempty', dTrees); 
    dTrees(emptyInds) = []; 
end